/**
 *  SatNet
 *
 *  Copyright (C) 2005  Yves Radunz, Volker Grabsch
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

import java.lang.*;
import java.util.*;
public class SatNet
{
    private Vector psat=new Vector();
    private double radius, masse, sathoch, satalarm;
    private int lengthquality=0;
    private double min, max, avr;

    /**
     * neues Satellitennetz für die Erde erzeugen
     */
    public SatNet(double hoch, double alarm)
    {
	this(hoch, alarm, 5.97e24, 6.37e6);
    }

    /**
     * neues Netz für einen anderen Planeten erstellen
     */
    public SatNet(double hoch, double alarm, double plmasse, double plradius)
    {
	sathoch=hoch; satalarm=alarm;
	masse=plmasse; radius=plradius;
    }

    /*****************************************************
     *
     *  Satelliten verwalten
     *
     *****************************************************/

    /**
     * Erzeugt einen neuen Satelliten, der die angegebenen Punkte überfliegt.
     */
    public void newsat(double west1, double nord1, double west2, double nord2)
    {
	psat.add(new Satellit(west1, nord1, west2, nord2, masse, sathoch, satalarm));
	lengthquality=0;
    }

    /**
     * neuen Satelliten mit einer zufälligen Flugbahn erstellen
     */
    public void newsatrnd()
    {
	boolean nochmal=true;
	while(nochmal){
	    try
		{
		    newsat(Math.random()*Math.PI*2, (2*Math.random()-1)*Math.PI, Math.random()*Math.PI*2, (2*Math.random()-1)*Math.PI);
		    nochmal=false;
		}
	    catch (IllegalArgumentException e) {nochmal=true;}
	}
    }

    /**
     * neuen Satelliten erstellen, der möglichst weit von den anderen Satelliten entfernt ist
     */
    public void newsat()
    {
	switch(size())
	    {
	    case 0:{
		// Der erste Satellit erhält eine beliebige Umlaufbahn.
		newsat(0,.4,Math.PI/2,0);
		break;
	    }
	    case 1:{
		// Der zweite Satellit befindet sich immer genau gegenüber dem ersten Satelliten
		double[] sat1=Vektor.polarinv(sat(0).getStart());
		double[] sat2=Vektor.polarinv(sat(0).getRichtung());
		newsat(sat1[1]+Math.PI, -sat1[2], sat2[1]+Math.PI, -sat2[2]);
		break;
	    }
	    default:{
		double s=Math.PI/Math.sqrt(size())/2;

		// Startpunkt als einen Punkt bestimmen, der von den anderen Satelliten zum Zeitpunkt 0 möglichst weit entfernt ist.
		double max=-1;
		double maxtheta=0, maxphi=0;
		Vektor startpunkt=new Vektor();
		for (double phi=0; phi<2*Math.PI;phi+=s)
		    for (double theta=-Math.PI/2; theta<Math.PI/2+.0001; theta+=s){
			Vektor punkt=Vektor.polar(radius,phi,theta);
			double d=mindist(punkt);
			if (max<d){
			    max=d;
			    startpunkt.set(punkt);
			    maxtheta=theta;
			    maxphi=phi;
			}
		    }

		// Zweiten Punkt als einen Punkt bestimmen, der einen möglichst großen Abstand zu den anderen Satelliten ermöglicht.
		max=-1;
		double maxalpha=0;
		if (Math.abs(maxtheta)<Math.PI/4){
		    for (double alpha=0; alpha<2*Math.PI; alpha+=s){
			newsat(maxphi, maxtheta, maxphi+Math.PI/2, alpha);
			double m=minsatdist();
			if (max<m){
			    max=m;
			    maxalpha=alpha;
			}
			removesat(size()-1);
		    }
		    newsat(maxphi, maxtheta, maxphi+Math.PI/2,maxalpha);
		}else{
		    for (double alpha=0; alpha<2*Math.PI; alpha+=s){
			newsat(maxphi, maxtheta, alpha, 0);
			double m=minsatdist();
			if (max<m){
			    max=m;
			    maxalpha=alpha;
			}
			removesat();
		    }
		    newsat(maxphi, maxtheta, maxalpha, 0);
		}
	    }
	    }
    }

    /**
     * Entfernt den Satelliten mit dem angegebenen Index.
     */
    public void removesat(int index)
    {
        psat.remove(index);
        lengthquality=0;
    }

    /**
     * Entfernt den zuletzt hinzugefügten Satelliten.
     */
    public void removesat()
    {
	try
	    {
		removesat(size()-1);
	    }
	catch (ArrayIndexOutOfBoundsException e) {}
    }

    /**
     * Gibt den Satelliten mit dem Index (0...size-1) zurück
     */
    public Satellit sat(int index)
    {
	return (Satellit) psat.get(index);
    }

    /**
     * Anzahl der Satelliten
     */
    public int size()
    {
	return psat.size();
    }


    /*****************************************************
     *
     *  Kollisionsgefahr
     *
     *****************************************************/

    /**
     * Abstand des Punktes zum nächsten Satelliten zum Zeitpunkt 0
     */
    private double mindist(Vektor punkt)
    {
	double min=Double.MAX_VALUE;
	for (int i=0; i<size(); i++)
	    min=Math.min(min,Vektor.dist(punkt, sat(i).getPosition(0)));
	return min;
    }

    /**
     * Minimaler Abstand der Satelliten
     */
    public double minsatdist()
    {
	double min=Double.MAX_VALUE;
	for (int i=0; i<size()-1; i++){
	    for (int j=i+1; j<size(); j++){
		min=Math.min(min,sat(i).mindist(sat(j)));
	    }
	}
	return min;
    }

    /**
     * besteht Kollisionsgefahr irgendwelcher Satelliten?
     */
    public boolean kollgef()
    {
	for (int i=0; i<size()-1; i++){
	    for (int j=i+1; j<size(); j++){
		if(sat(i).kollgef(sat(j))) return true;
	    }
	}
	return false;
    }

    /**
     * Kollisionsmöglichkeiten zum Zeitpunkt zeit
     */
    public boolean[] kollisionen(double zeit)
    {
	//Welche Satelliten können zu diesem Zeitpunkt kolliedieren?
	boolean[] kollidiert=new boolean[size()];
	Arrays.fill(kollidiert,false);
	for (int i=0; i<size()-1; i++){
	    for (int j=i+1; j<size(); j++){
		if(sat(i).dist(sat(j),zeit)<satalarm){
		    kollidiert[i]=true;
		    kollidiert[j]=true;
		}
	    }
	}
	return kollidiert;
    }

    /*****************************************************
     *
     *  Kommunikationsnetz
     *
     *****************************************************/

    /**
     * führt die direkte Verbindung nicht durch den Planeten?
     */
    private boolean notcutsplanet(Vektor stuetz, Vektor richt)
    {
	if (richt.iszero()) {return true;}
	return Vektor.distGerade(stuetz,richt)>radius;
    }

    /**
     * neues Kommunikationsnetz zum Zeitpunkt zeit erstellen
     */
    public Graph createnet(double zeit)
    {
	// alle möglichen Verbindungen zwischen zwei Satelliten auflisten,
	// die nicht den Planeten schneiden
	Graph graph=new Graph();
	for (int i=0; i<size()-1; i++)
	    for (int j=i+1; j<size(); j++)
		if(notcutsplanet(sat(i).getPosition(zeit),Vektor.sub(sat(i).getPosition(zeit),sat(j).getPosition(zeit))))
		    graph.addedge(i,j);
	graph.optimize(size(),psat,zeit);
	return graph;
    }

    /**
     * minimale, durchschnittliche und maximale Netzlaenge berechnen
     */
    private void findlengths(int quality)
    {
        if (quality<=0 || size()==0){
	    min=0; max=0; avr=0;
	}
        double maxzeit=sat(0).getUzeit(),
	    schritt=maxzeit/quality;
        min=sathoch*2*size();
        max=0;
	avr=0;
        for (double i=0; i<maxzeit; i+=schritt){
	    Graph graph=createnet(i);
	    double laenge=graph.getNetLength();
	    min = Math.min(min, laenge);
	    max = Math.max(max, laenge);
	    avr += laenge;
	}
	avr /= quality;
	lengthquality = quality;
    }
    
    /**
     * minimale Netzlaenge mit einer bestimmten Qualitaet bestimmen
     */
    public double minlength(int quality)
    {
        if (quality!=lengthquality)
            findlengths(quality);
        return min;
    }

    /**
     * Standardqualität der Längen festlegen
     */
    public int stdquality()
    {
	return size()<100?10000/size():100;
    }

    /**
     * minimale Netzlaenge bestimmen
     */
    public double minlength()
    {
        return minlength(stdquality());
    }

    /**
     * maximale Netzlaenge mit einer bestimmten Qualitaet bestimmen
     */
    public double maxlength(int quality)
    {
        if (quality!=lengthquality)
            findlengths(quality);
        return max;
    }

    /**
     * maximale Netzlaenge bestimmen
     */
    public double maxlength()
    {
        return maxlength(stdquality());
    }

    /**
     * durchschnittliche Netzlaenge mit einer bestimmten Qualitaet bestimmen
     */
    public double avrlength(int quality)
    {
        if (quality!=lengthquality)
            findlengths(quality);
        return avr;
    }

    /**
     * durchschnittliche Netzlaenge bestimmen
     */
    public double avrlength()
    {
        return avrlength(stdquality());
    }
}
